import hashlib
import json

from flask import current_app
from psycopg2.extras import NumericRange

from stumpf.diva_fact_schemas import ActivityFactPropertySchema, \
    TrackDetectionFactPropertySchema, \
    TrackFactPropertySchema
from stumpf.models import db


class BaseActivityGroupImporter(object):

    def import_activity_group(self, activity_group, video, provenance=None):
        provenance = dict(provenance)
        assert len(activity_group['detections']) == 2

        frame_start = self.detection_frame(activity_group['detections'][0])
        frame_end = self.detection_frame(activity_group['detections'][1]) + 1

        data = json.dumps(activity_group, sort_keys=True).encode()
        activity_group_hash = hashlib.md5(data).hexdigest()
        provenance['activity_group_json_hash'] = activity_group_hash

        activity, activity_provenance = video.get_fact_from_external_process(
            self.activity_schema.load({
                'type': activity_group['type']
            }),
            frame_range=(frame_start, frame_end),
            schema=self.activity_schema_name,
            provenance=provenance
        )
        db.session.add_all([activity, activity_provenance])

        tracks = []
        for track_object in activity_group['tracks']:
            tracks.append(self.import_track(track_object, activity, video, provenance))

        # make sure the frame ranges are consistent
        track_lower_frame = min([track.frame_range.lower for track in tracks])
        track_upper_frame = max([track.frame_range.upper for track in tracks])
        try:
            assert track_lower_frame == activity.frame_range.lower
            assert track_upper_frame == activity.frame_range.upper
        except Exception:
            # When the assertion fails use the track frame range over the frame range given
            # by the activity object.
            current_app.logger.warn(
                f'Activity frame range ({activity.frame_range.lower},{activity.frame_range.upper}) '
                f'did not match track frame range ({track_lower_frame},{track_upper_frame}) '
                f'for "{activity.id}".'
            )
            activity.frame_range = NumericRange(track_lower_frame, track_upper_frame, bounds='[]')

        return activity

    def import_track(self, track_object, activity, video, provenance=None):
        provenance = dict(provenance)
        track_type = track_object['type']
        frames = [self.detection_frame(detection) for detection in track_object['detections']]
        frame_start = min(frames)
        frame_end = max(frames) + 1
        track, track_provenance = video.get_fact_from_external_process(
            self.track_schema.load({
                'activity_id': str(activity.id),
                'type': track_type
            }),
            frame_range=(frame_start, frame_end),
            schema=self.track_schema_name,
            provenance=provenance
        )
        db.session.add_all([track, track_provenance])

        for detection_object in track_object['detections']:
            self.import_detection(detection_object, track, activity, video, provenance)
        return track

    def import_detection(self, detection_object, track, activity, video, provenance=None):
        provenance = dict(provenance)
        detection_frame = self.detection_frame(detection_object)
        detection, detection_provenance = video.get_fact_from_external_process(
            self.detection_schema.load({
                'type': track.fact['type'],
                'activity_id': str(activity.id),
                'track_id': str(track.id),
                'source': self.detection_object_source(detection_object),
                'occlusion': 'none',
                **self.bounding_box(detection_object)
            }),
            frame_range=(detection_frame, detection_frame + 1),
            schema=self.detection_schema_name,
            provenance=provenance
        )
        db.session.add_all([detection, detection_provenance])
        return detection

    # for backward compatibility with old files
    def detection_object_source(self, detection):
        return detection.get('src') or detection.get('source')

    def detection_frame(self, detection):
        return detection.get('ts0', detection.get('frame'))

    def bounding_box(self, detection):
        if 'g0' in detection:
            bbox = [int(pt) for pt in detection['g0'].split(' ')]
            return {
                'left': bbox[0],
                'right': bbox[2],
                'top': bbox[1],
                'bottom': bbox[3]
            }
        return {
            'left': detection['left'],
            'right': detection['right'],
            'top': detection['top'],
            'bottom': detection['bottom']
        }


class GrozperActivityGroupImporter(BaseActivityGroupImporter):

    activity_schema_name = 'grozper:activity'
    activity_schema = ActivityFactPropertySchema()
    track_schema_name = 'grozper:track'
    track_schema = TrackFactPropertySchema()
    detection_schema_name = 'grozper:track_detection'
    detection_schema = TrackDetectionFactPropertySchema()


grozper_activity_importer = GrozperActivityGroupImporter()
