from glob import glob
import json
import os
import sys


def read_track_types(dirname):
    # type_id -> track type
    type_file = glob(os.path.join(dirname, '*.kw18.types'))
    assert len(type_file) == 1

    types = {}
    with open(type_file[0]) as f:
        for line in f.readlines():
            if line[0] == '#':
                continue
            row = line.strip().split(' ')
            types[int(row[0])] = row[1]
    return types


def read_detections(dirname):
    # track_id -> ( frame_id -> bounding_box )
    detection_file = glob(os.path.join(dirname, '*.kw18'))
    assert len(detection_file) == 1

    detections = {}
    with open(detection_file[0], 'r') as f:
        for line in f.readlines():
            if line[0] == '#':
                continue
            row = line.strip().split(' ')
            id_ = int(row[0])
            frame = int(row[2])
            bbox = [round(float(v)) for v in row[9:13]]
            track = detections.setdefault(id_, {})
            track[frame] = bbox

    return detections


def fetch_track_detections(detections):
    track_detections = []
    for frame, bbox in detections.items():
        track_detections.append({
            'frame': frame,
            'source': 'ground-truth',
            'left': bbox[0],
            'right': bbox[2],
            'top': bbox[1],
            'bottom': bbox[3]
        })
    assert len(track_detections) >= 2
    return track_detections


def read_activity(line, detections, track_types):
    # from reverse engineering:
    ACTIVITY_ID_IDX = 1
    START_FRAME_IDX = 4
    END_FRAME_IDX = 6
    NUM_TRACKS_IDX = 12
    TRACK_ID_IDX = 13

    row = line.strip().split(' ')
    activity_id = int(row[ACTIVITY_ID_IDX])
    start_frame = int(row[START_FRAME_IDX])
    end_frame = int(row[END_FRAME_IDX])
    number_of_tracks = int(row[NUM_TRACKS_IDX])

    if number_of_tracks == 1:
        track_id = int(row[TRACK_ID_IDX])
        track_type = track_types[track_id]
        tracks = [{
            'type': track_type,
            'detections': fetch_track_detections(detections[track_id])
        }]
    else:
        tracks = []
        for i in range(number_of_tracks):
            track_id = int(row[TRACK_ID_IDX + i])
            track_type = track_types[track_id]

            tracks.append({
                'type': track_type,
                'detections': fetch_track_detections(
                    detections[track_id])
            })

    return {
        'type': ACTIVITY_TYPE_MAP[activity_id],
        'detections': [{
            'source': 'ground-truth',
            'frame': start_frame
        }, {
            'source': 'ground-truth',
            'frame': end_frame
        }],
        'tracks': tracks
    }


ACTIVITY_TYPE_MAP = {
    2: 'Vehicle_Moving',
    3: 'Stand_Up',
    4: 'Sit_Down',
    5: 'Carry_in_Hands',
    6: 'Carry_on_Back',
    7: 'Talk_on_Phone',
    8: 'Read_Document',
    9: 'Text_On_Phone',
    10: 'Vehicle_Stopping',
    11: 'Vehicle_Starting',
    12: 'Vehicle_Turning_Left',
    13: 'Vehicle_Turning_Right',
    14: 'Vehicle_UTurn',
    15: 'Object_Transfer',
    16: 'Pick_Up_Object',
    17: 'Set_Down_Object',
    18: 'Heavy_Carry',
    19: 'Hand_Interaction',
    20: 'Embrace_Interaction',
    21: 'Open_Facility_Door',
    22: 'Close_Facility_Door',
    23: 'Enter_Facility',
    24: 'Exit_Facility',
    25: 'Open_Vehicle_Door',
    26: 'Close_Vehicle_Door',
    27: 'Enter_Vehicle',
    28: 'Exit_Vehicle',
    29: 'Open_Trunk',
    30: 'Close_Trunk',
    31: 'Load_Vehicle',
    32: 'Unload_Vehicle',
    33: 'Vehicle_PicksUp_Person',
    34: 'Vehicle_DropsOff_Person',
    35: 'People_Talking',
    36: 'Joining_Queue',
    37: 'Riding',
    38: 'Purchasing',
    39: 'Laptop_Interaction',
    40: 'Abandon_Package',
    41: 'Casing_Facility',
    43: 'Theft',
    44: 'Vehicle_Reversing'
}


def read_all_activities(dirname):
    detections = read_detections(dirname)
    track_types = read_track_types(dirname)
    activities_file = glob(os.path.join(dirname, '*.txt'))
    assert len(activities_file) == 1

    activities = []
    with open(activities_file[0], 'r') as f:
        for line in f.readlines():
            if line[0] == '#':
                continue
            activities.append(read_activity(line, detections, track_types))
    return activities


def kw2json(project_path, kw18_fbase):
    for activity in read_all_activities(project_path):
        yield activity


if __name__ == '__main__':
    print(json.dumps(read_all_activities(sys.argv[1])))
