import munkres

from stumpf.diva_fact_schemas import ActivityFact, ActivityFactSchema
from stumpf.models import Assignment, db, Task

activity_fact_schema = ActivityFactSchema()
MINIMUM_OVERLAP = 0.5
GROUND_TRUTH_QUERY = """
select b.id as activity_id,
       a.id as truth_id,
       a.video_id,
       a.frame_range,
       a.fact->>'type' as type
from fact a
join fact b
    on a.video_id = b.video_id
        and a.fact->>'type' = b.fact->>'type'
        and lower(a.frame_range) = lower(b.frame_range)
        and upper(a.frame_range) = upper(b.frame_range)
        and a.id <> b.id
join provenance p
    on a.id = p.fact_id
where p.meta->>'refiner-ground-truth' is not null
    and a.schema = 'grozper:activity'
    and b.id = '%s'
order by
    a.created desc
"""


def app_context():
    from stumpf.app import create_app
    app = create_app({'SQLALCHEMY_ECHO': False})
    return app.app_context()


def get_ground_truth(assignment):
    task = Task.query.get_or_404(assignment.task_id)
    query = db.session.execute(GROUND_TRUTH_QUERY % task.args['activityId'])
    truth = query.first()
    assert truth is not None, 'Ground truth not found for the provided activity'

    activity_id = truth.truth_id
    return activity_fact_schema.dump(ActivityFact.query.get(activity_id))


def get_assignments(task_id):
    query = Assignment.query
    query = query.filter_by(task_id=task_id)
    query = query.filter(Assignment.status.in_(['Approved', 'Submitted']))
    query = query.filter(Assignment.answer != None)  # noqa
    return query.all()


def convert_answer_detection_to_fact_object(detection):
    bbox = detection['bbox']
    return {
        'frame': detection['frame'],
        'left': bbox[0],
        'right': bbox[2],
        'top': bbox[1],
        'bottom': bbox[3],
        'source': detection['src']
    }


def generate_answer_tracks(answer):
    tracks = []
    for detection in answer['detections']:
        id_ = detection['id1'] - 1
        if id_ >= len(tracks):
            tracks.append({
                'detections': [],
                'keyframes': []
            })
        converted_detection = convert_answer_detection_to_fact_object(detection)
        tracks[id_]['detections'].append(converted_detection)
        if converted_detection['source'] == 'truth':
            tracks[id_]['keyframes'].append(converted_detection)

    # sort the detection lists
    for track in tracks:
        track['detections'] = sorted(track['detections'], key=lambda d: d['frame'])
        track['keyframes'] = sorted(track['keyframes'], key=lambda d: d['frame'])

    return tracks


def frame_matches(frame1, frame2):
    return frame1['frame'] == frame2['frame'] and \
        frame1['left'] == frame2['left'] and \
        frame1['right'] == frame2['right'] and \
        frame1['top'] == frame2['top'] and \
        frame1['bottom'] == frame2['bottom']


def detection_area(detection):
    if 'area' not in detection:
        detection['area'] = \
            (detection['right'] - detection['left']) * (detection['bottom'] - detection['top'])
    return detection['area']


def intersection(detection1, detection2):
    left = max(detection1['left'], detection2['left'])
    right = min(detection1['right'], detection2['right'])
    top = max(detection1['top'], detection2['top'])
    bottom = min(detection1['bottom'], detection2['bottom'])
    return {
        'left': left,
        'right': max(left, right),
        'top': top,
        'bottom': max(top, bottom)
    }


def intersection_over_union(detection1, detection2):
    intersection_area = detection_area(intersection(detection1, detection2))
    union_area = detection_area(detection1) + detection_area(detection2) - intersection_area
    if union_area == 0:
        return 1
    return intersection_area / union_area


def score_track(answer_track, detections):
    total_score = 0
    number_of_frames = len(detections)
    assert number_of_frames == len(answer_track['detections'])

    for i in range(number_of_frames):
        # for iou metric:
        # total_score += intersection_over_union(detections[i], answer_track['detections'][i])
        # for grouper paper metric:
        total_score += total_grouper_score(detections[i], answer_track['detections'][i])

    return total_score / number_of_frames


def get_max_score_from_matrix(score_matrix):
    """
    Uses the hungarian algorithm to get the optimal pairing between
    ground truth tracks and answer tracks.
    """
    m = munkres.Munkres()

    # turn the maximization problem into a minimization problem over integers
    cost_matrix = munkres.make_cost_matrix(score_matrix, lambda x: int(1000 * (1 - x)))

    # compute the optimal path and compute the score
    optimal_path = m.compute(cost_matrix)
    value = 0
    for i, j in optimal_path:
        value += score_matrix[i][j]
    return value / len(optimal_path)


def score_assignment(assignment):
    truth = get_ground_truth(assignment)
    answer_tracks = generate_answer_tracks(assignment.answer)

    # score all possible pairings of answer tracks to truth tracks
    track_score_matrix = []
    for truth_track in truth['tracks']:
        track_scores = []
        detections = sorted(truth_track['detections'], key=lambda d: d['frame'])
        for answer_track in answer_tracks:
            if len(detections) == len(answer_track['detections']):
                score = score_track(answer_track, detections)
            else:
                score = 0
            track_scores.append(score)

        track_score_matrix.append(track_scores)

    return get_max_score_from_matrix(track_score_matrix)


# Metrics described by:
# https://goo.gl/1KEpGG
def overlap_score(box1, box2):
    area1 = detection_area(box1)
    area2 = detection_area(box2)
    largest_area = max(area1, area2)
    overlap_area = detection_area(intersection(box1, box2))
    if largest_area == 0:
        return 0
    return overlap_area / largest_area


def size_score(box1, box2):
    area1 = detection_area(box1)
    area2 = detection_area(box2)
    min_area = min(area1, area2)
    max_area = max(area1, area2)
    return 1 - (1 - min_area / max_area) / (1 - MINIMUM_OVERLAP)


def box_ratio(box):
    return (box['right'] - box['left']) / (box['bottom'] - box['top'])


def shape_score(box1, box2):
    ratio1 = box_ratio(box1)
    ratio2 = box_ratio(box2)
    min_ratio = min(ratio1, ratio2)
    max_ratio = max(ratio1, ratio2)
    return 1 - (1 - min_ratio / max_ratio) / (1 - MINIMUM_OVERLAP * MINIMUM_OVERLAP)


def box_center(box):
    return (box['left'] + box['right']) / 2, (box['top'] + box['bottom']) / 2


def position_score(box1, box2):
    center_x1, center_y1 = box_center(box1)
    center_x2, center_y2 = box_center(box2)
    delta_x = abs(center_x1 - center_x2)
    delta_y = abs(center_y1 - center_y2)
    max_width = max(box1['right'] - box1['left'], box2['right'] - box2['left'])
    max_height = max(box1['bottom'] - box1['top'], box2['bottom'] - box2['top'])
    distance_width_ratio = delta_x / max_width
    distance_height_ratio = delta_y / max_height
    return 1 - (distance_width_ratio + distance_height_ratio) / (2 * (1 - MINIMUM_OVERLAP))


def total_grouper_score(box1, box2):
    overlap = overlap_score(box1, box2)
    if overlap < MINIMUM_OVERLAP:
        return overlap / 4

    size = size_score(box1, box2)
    shape = shape_score(box1, box2)
    position = position_score(box1, box2)
    return (overlap + size + shape + position) / 4


def main(task_id):
    with app_context():
        for assignment in get_assignments(task_id):
            score = score_assignment(assignment)
            url = f'https://stumpf.kitware.com/#/refiner/answer/{str(assignment.id)}'
            print('%6.4f - %s' % (score, url))


if __name__ == '__main__':
    import sys
    main(sys.argv[1])
