import json
import logging
import os
from shutil import copyfileobj
import time
from urllib.request import urlopen
from sqlalchemy import cast
import sqlalchemy

import click
import hvac
from stumpf.app import create_app
from stumpf.models import db, Provenance
from stumpf.diva_fact_schemas import ActivityFact, GladiatorFact

from stumpfwald import authorized, cli
from stumpfwald.refiner import preview

REFINER_APP_ID = 'a05be309-1999-45ac-845d-f66eb76ebeba'


def get_production_database_uri(vault_token):
    client = hvac.Client(url='https://vault.kitware.com', token=vault_token)
    db_creds = client.read('secret/data/projects/diva-mturk/stumpf/rds')['data']['data']
    return f"postgresql://{db_creds['username']}:{db_creds['password']}@{db_creds['host']}/{db_creds['dbname']}"  # noqa


def download_video(video):
    path = video.path
    fname = path.split('/')[-1]
    if not os.path.exists(fname):
        click.echo('Downloading video...', err=True)
        with urlopen(path) as u, open(fname, 'wb') as f:
            copyfileobj(u, f)
    return fname


@click.group(name='refiner', short_help='The Grozper related commands.')
@click.option('--production', is_flag=True, default=False,
              help='Use the production Stumpf database.')
@click.option('--vault-token', type=click.UUID, help='The token to use to authenticate to Vault.')
@click.pass_obj
def refiner(ctx, production, vault_token):
    logging.getLogger('flask.app').setLevel(logging.ERROR)
    ctx['production'] = production
    app_config = {
        'SQLALCHEMY_ECHO': False,
    }
    if ctx['production']:
        app_config['SQLALCHEMY_DATABASE_URI'] = get_production_database_uri(vault_token)

    ctx['app'] = create_app(app_config)


@refiner.command(name='task-answer', help='Generate a video clip of a Refiner fact')
@click.argument('task-id', type=click.UUID)
@click.argument('output', required=False)
@click.pass_obj
def generate_task_answer(ctx, task_id, output):
    with ctx['app'].app_context():
        task_id = str(task_id)
        task = preview.get_refiner_task(task_id)
        output = output or f'task_{task_id}.mp4'
        video = download_video(task.video)
        click.echo('Generating ffmpeg filters...', err=True)
        detections = preview.get_refiner_fact_detections(task_id)
        click.echo('Generating preview...', err=True)
        preview.generate_clip(detections, video, output)


@refiner.command(name='assignment-answer',
                 help='Generate a video clip of a Refiner assignment answer')
@click.argument('assignment-id', type=click.UUID)
@click.argument('output', required=False)
@click.pass_obj
def generate_assignment_answer(ctx, assignment_id, output):
    with ctx['app'].app_context():
        assignment_id = str(assignment_id)
        assignment = preview.get_refiner_assignment(assignment_id)
        if assignment.answer is None:
            raise Exception('No answer exists for the given assignment')
        task = assignment.task
        output = output or f'assignment_{assignment_id}.mp4'
        video = download_video(task.video)
        click.echo('Generating ffmpeg filters...', err=True)
        detections = preview.get_refiner_answer_detections(assignment)
        click.echo('Generating preview...', err=True)
        preview.generate_clip(detections, video, output)


@refiner.command(name='hit-from-gladiator', help='Create Refiner HITs from Gladiator fact')
@click.option('--after', type=click.DateTime())
@click.option('--before', type=click.DateTime())
@click.option('--grozper-batch', type=click.STRING)
@click.option('--sandbox', is_flag=True)
@click.option('--include-good', is_flag=True)
@click.option('--dry-run', is_flag=True)
@click.pass_obj
@authorized
def create_hits_from_gladiator_fact_(ctx, after, before, grozper_batch, sandbox, include_good, dry_run):
    if not after and not grozper_batch:
        raise click.UsageError('--after or --grozper-batch is required')

    with ctx['app'].app_context():
        if after:
            query = GladiatorFact.query.filter(GladiatorFact.created >= after)
            if before:
                query = query.filter(GladiatorFact.created < before)
        elif grozper_batch:
            sub = db.session.query(Provenance.fact_id).join(
                ActivityFact).filter(Provenance.meta['batch'].astext == grozper_batch)
            # Have to cast to in memory tuple otherwise SQL complain about text and uuid comparison
            sub = (str(item[0]) for item in sub.all())
            query = GladiatorFact.query.filter(GladiatorFact.fact['activity_id'].astext.in_(sub))
        if not include_good:
            query = query.filter(GladiatorFact.fact['good'].astext == 'false')

        if dry_run:
            click.echo(f'{query.count()} Refiner HITs would be created')
            return

        for gladiatorFact in query:
            r = ctx['session'].post(f'task/{gladiatorFact.video_id}/{REFINER_APP_ID}/hit', json={
                'sandbox': sandbox,
                'title': 'Video annotation',
                'description': 'Improve object annotation',
                'assignment_duration_in_seconds': 3600,
                'max_assignments': 5,
                'reward':  20,
                'lifetime_in_seconds': 24 * 3600,
                'args': {
                    'activtyId': gladiatorFact.fact['activity_id']
                }
            })
            if not r.ok:
                click.echo(r.text, err=True)
            else:
                click.echo(json.dumps(r.json(), indent=4))
            time.sleep(2)


cli.add_command(refiner)
